import numpy as np
import pandas as pd

class Helpers():
    
    @staticmethod
    def docs():
        print('Scikit learn RF hyperparameters')
        print(' - https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html')
        print('\n')

        print('XGBoost classifier hyperparameters')
        print(' - https://xgboost.readthedocs.io/en/latest/parameter.html')
        print('\n')

    @staticmethod
    def calc_params(x):
        n = 1/1000000
        for i in [len(v) for k,v in x.items()]:
            n = n*i
        return n

    @staticmethod
    def grid_sklearn_RFC():
        
        grid={
            'criterion':['entropy', 'gini'],
            'max_depth':[None,2,4,6,8], 
            'max_features':['sqrt', 'log2','auto',0.95, 0.90, 0.85, 0.80, 0.75, 0.70],
            'max_leaf_nodes':[None,10, 15, 20, 25, 30, 35, 40, 45, 50],
            'max_samples':[0.5, 0.6, 0.7, 0.8, 0.9, None],
            'min_impurity_decrease':[0.0, 0.0005, 0.005, 0.05, 0.10, 0.15, 0.2],
            'min_samples_leaf':[1, 0.01, 0.02, 0.03, 0.04], 
            'min_samples_split':[2, 3, 4, 5, 6, 8, 10],
            'min_weight_fraction_leaf':[0.0, 0.0025, 0.005, 0.0075, 0.01, 0.05],
            'n_estimators':[10, 40, 100]
        }

        n = Helpers.calc_params(grid)        
        print(f'Note, grid contains {np.round(n,1)} mn combinations!')
        print('Check docs for details on hyperparameters: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html')
        print('\n')

        return grid

    @staticmethod
    def grid_xgboost_XGBC(booster='gbtree'):
        
        # gbtree default
        grid = {
            'colsample_bylevel':[0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            'colsample_bynode':[0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            'colsample_bytree':[0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            'gamma':[0, 0.01, 0.05, 0.1, 0.5, 1, 2, 3],
            'learning_rate':[0.1, 0.2, 0.3, 0.4, 0.5],
            'max_depth':[1, 2, 3, 4, 5, None],
            'min_child_weight':[1, 2, 3, 4, 5],
            'n_estimators':[2, 25, 50, 75, 100],
            'num_parrallal_tree':[1,10,20],
            'subsample':[0.5, 0.6, 0.7, 0.8, 0.9, 1],
        }

        if booster == 'dart':
            grid['sample_type']=['uniform', 'weighted']
            grid['normalize_type']=['tree', 'forest']
            grid['rate_drop']=[0, 0.1, 0.2, 0.3, None]
            grid['one_drop']=[0, 1]
            grid['skip_drop']=[0, 0.25, 0.5, 0.75, None]

        elif booster == 'gblinear':
            grid = {
                'reg_alpha':[0.0,0.1,0.5,1.0],  #L1 regularization
                'reg_lambda':[0.0,0.1,0.5,1.0],  #L2 regularization
                'colsample_bylevel':[0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                'colsample_bynode':[0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                'colsample_bytree':[0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                'learning_rate':[0.1, 0.2, 0.3, 0.4, 0.5],
                'n_estimators':[2, 25, 50, 75, 100],
            }

        
        n = Helpers.calc_params(grid)   
        print(f'Note, grid contains {np.round(n,1)} mn combinations!')
        print('\n')

        return grid



    @staticmethod
    def list_transformers():
        transformers =  {
            'general':[
                ('ColSelector', 'Select column in df'),
                ('PipeMerger', 'Combine pipelines'),
                ('TypeCaster', 'Cast col to type'),
            ],
            'handle nan & outliers':[
                ('NullishToNan', 'Replace nullish vals with np.nan'),
                ('NanImputer', 'NaN to imputed value'),
                ('OutlierTrimmer', 'Trim outliers'),
            ],
            'scalers & transformers':[
                ('StdScaler', 'Standard scale feature'),
                ('MinMaxScaler', 'Min-max scale feature'),
                ('BoxCoxTransformer', 'Box-cox transform feature')
            ],
            'encoders':[
                ('OrdinalEncoder', 'Ordinal encode catcol'),
                ('OnehotEncoder', 'Onehot encode'),
            ]
        }
        print('TRANSFORMERS IN PACKAGE:\n')
        for k,v in transformers.items():
            print(f'\n{k.upper()} --->')
            for e in v:
                print(f'{e[0]}:\t\t{e[1]}')




import numpy as np
import pandas as pd
from typing import List, Dict, Union

class DescStat():

    NULLISH=[np.nan, 0, -999, 'missing', '?', '', ' ', '  ', '   ']

    @staticmethod
    def correl(df:pd.DataFrame, x:str,y:str)->float:
        try:
            return np.corrcoef(df[x],df[y])[0,1]
        except:
            return np.nan


    @staticmethod
    def catcol(df:pd.DataFrame, col:str)->pd.DataFrame:
        '''
        Generate stats for single categorical column in dataframe
        '''
        try:
            return pd.DataFrame({
                'n_obs':df[col].value_counts(),
                'freq':df[col].value_counts(normalize=True),
                'n_missing':df.groupby(col)[col].count().rsub(df.groupby(col)[col].size(), axis=0),
                'pct_tot_missing': df.groupby(col)[col].count().rsub(df.groupby(col)[col].size(), axis=0)/df.shape[0]
            }).T  
        except:
            print('Column must be categorical')

    @staticmethod
    def catcol_trgt(df:pd.DataFrame, col:str, response:str)->pd.DataFrame:
        '''
        Generate stats for single categorical column in dataframe.
        Compare with response
        '''
        try:
            return pd.DataFrame({
                'mean_trgt':df.groupby(col)[response].mean(),
                'n_obs':df[col].value_counts(),
                'freq':df[col].value_counts(normalize=True),
                'n_missing':df.groupby(col)[col].count().rsub(df.groupby(col)[col].size(), axis=0),
                'pct_tot_missing': df.groupby(col)[col].count().rsub(df.groupby(col)[col].size(), axis=0)/df.shape[0]
            }).T        
        except:
            print('Catcol must be categorical, target must be numeric')

    @staticmethod
    def numcol(df:pd.DataFrame, col:str, nullish:List=NULLISH)->pd.DataFrame:
        '''
        Generate stats for single numerical column in dataframe
        '''

        try:
            q25 = df[col].quantile(.25)
            q75 = df[col].quantile(.75)
            IQR = q75-q25
            stats_frame =  pd.DataFrame({            
                'mean':[df[col].mean()],
                'min':[df[col].min()],
                'p25':[df[col].quantile(.25)],
                'p50':[df[col].quantile(.5)],
                'p75':[df[col].quantile(.75)],
                'max':[df[col].max()],
                'skew(pos=right tail)':[df[col].skew()],
                'fisher_kurtosis(pos=thick tail)':[df[col].kurt()],
                'neg_outlier_threshold':[q25-1.5*IQR],
                'neg_outliers': [df.loc[df[col]<(q25-1.5*IQR),col].shape[0]],
                'pos_outlier_threshold':[q75+1.5*IQR],
                'pos_outliers': [df.loc[df[col]>(q75+1.5*IQR),col].shape[0]],
                'n_nan': [df[col].isin([np.nan]).sum()],
                'n_nullish': [df[col].isin(nullish).sum()],
                'pct_tot_nullish': [df[col].isin(nullish).sum()/df.shape[0]],
                'dtype':[df[col].dtype]
            }).T
            stats_frame.columns=[col]
            return stats_frame
        except:
            print('Column must be numeric')

    @staticmethod
    def numcol_trgt(df:pd.DataFrame, col:str, response:str, nullish:List=NULLISH)->pd.DataFrame:
        '''
        Generate stats for single numerical column in dataframe
        '''
        try:
            stats_frame=DescStat.numcol(df, col)
            response_frame=pd.DataFrame({
                'correl_trgt':[DescStat.correl(df, col, response)],
                'std':[df[col].std()],
            }).T
            response_frame.columns=[col]
            return pd.concat([response_frame, stats_frame], axis=0)
        except:            
            print('Both columns must be numeric')


    @staticmethod
    def df(df:pd.DataFrame, nullish:List=NULLISH, \
            trgt=None, page_num:int=None, page_size:int=4)->pd.DataFrame:
        '''
        Generate statistics for pandas dataframe
        '''
        n_pages=1
        if page_num is not None:
            cols=df.columns
            n_cols=len(cols)
            n_pages=int(np.ceil(n_cols/page_size))
            start = page_num*page_size
            df = df.iloc[:,start:start+page_size]            

        corr_series = {c:'no_trgt' for c in df.columns}
        if trgt is not None:
            corr_series = pd.Series({c:DescStat.correl(df,trgt,c) for c in df.columns})

        q75=df.quantile(.75)
        q25=df.quantile(.25)
        IQR = q75-q25
        q_names = IQR.index

        return pd.DataFrame({
            'dtype':df.dtypes,
            'mode':df.mode().iloc[0,:],
            'cardinality':df.nunique(),
            'n_nan':df.isin([np.nan]).sum(),
            'n_nullish':df.isin(nullish).sum(),
            'corr_trgt':corr_series,
            'mean':df.mean(),
            'std':df.std(),
            'kurt(pos=thick tail)':df.kurt(),
            'skew(pos=right tail)':df.skew(),    
            'min':df.min(),
            'p_05':df.quantile(.05, axis=0),
            'p_25':df.quantile(.25, axis=0),        
            'p_50':df.quantile(.50, axis=0),
            'p_75':df.quantile(.75, axis=0),
            'p_95':df.quantile(.95, axis=0),
            'max':df.max(),
            'pos_outlier_thresh':q75+1.5*IQR,
            'neg_outlier_thresh':q25-1.5*IQR,
            'pos_outlier_count':(df[q_names] > q75+1.5*IQR).sum(),                    
            'neg_outlier_count':(df[q_names] < q25-1.5*IQR).sum(), 
            'page_displayed':np.where(page_num is not None, f'{page_num}/{n_pages-1}', 'all')
        }).T

    @staticmethod
    def col(df:pd.DataFrame,col:str, trgt=None, cardinality_bound:int=6):
        
        assert type(col) == str, "Function support only one column"
        assert col != trgt, "Col and Trgt can not be same value"
        df_slicer = [col,trgt] \
            if trgt is not None else [col]
        
        print(f'\nSUMMARY OF {col}:')
        print(DescStat.df(df[df_slicer], trgt=trgt).loc[:,col].dropna())

        if df[col].nunique() <= cardinality_bound:
            if not trgt:
                print('\n',DescStat.catcol(df, col))
            else:
                print('\n',DescStat.catcol_trgt(df, col, trgt))
        else:
            print(f'\nNote, cardinality table supressed for \
                cardinality greater than {cardinality_bound}. \
                Increase threshold to print')


from matplotlib.dates import DateFormatter, MonthLocator
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import shapiro, pearsonr, linregress
from scipy.stats.morestats import _calc_uniform_order_statistic_medians, _parse_dist_kw
import seaborn as sns
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.manifold import TSNE
import umap.umap_ as umap


sns.set_theme(
    style='darkgrid', 
    palette='magma', 
    font='sans-serif', 
    font_scale=1, 
    color_codes=True, 
    rc=None)


def ts_outlier_plot(data, dt, y):
    
    fig,ax=plt.subplots(figsize=(12,7))

    date_form = DateFormatter("%m-%d")
    ax.xaxis.set_major_formatter(date_form)

    mu=data[y].mean()
    std=data[y].std()
    q3=data[y].quantile(.75)
    q1=data[y].quantile(.25)
    iqr=q3-q1
    uthresh=q3+1.5*iqr
    lthresh=q1-1.5*iqr

    df_high = data[(data[y]<lthresh) | (data[y]>uthresh)]

    sns.lineplot(data=data, x=dt, y=y, ax=ax, zorder=1)
    sns.scatterplot(data=df_high, x=dt, y=y, ax=ax, color=sns.color_palette()[3], zorder=2, s=120, label='outliers')
    plt.xticks(rotation=40)

    # Ensure a major tick for each week using (interval=1) 
    ax.xaxis.set_major_locator(MonthLocator(interval=1))

    ax.axhline(uthresh, ls='--', alpha=.9, label='q3+1.5*iqr', color=sns.color_palette()[3])
    ax.axhline(lthresh, ls='--', alpha=.9, label='q1-1.5*iqr', color=sns.color_palette()[3])

    ax.axhline(mu, ls=':', alpha=.3, color=sns.color_palette()[0])
    ax.axhline(mu+std, ls=':', alpha=.5, label='mu+std', color=sns.color_palette()[1])
    ax.axhline(mu-std, ls=':', alpha=.7, label='mu-std', color=sns.color_palette()[0])

    ax.legend()
    ax.set_title(f'Outlier plot for feature {y}', fontweight='bold', fontsize=17)
    plt.tight_layout()



def separability_plot(df, trgt):
    df=df.copy()
    
    cls = df.pop(trgt)
    df=df.select_dtypes(include='number')

    fig,ax=plt.subplots(1,3,figsize=(15,5))

    # TSNE
    clf = TSNE(n_components=2, perplexity=30, learning_rate=15, random_state=1)
    X_t = clf.fit_transform(df)
    x_=pd.DataFrame(X_t,columns=['x','y'])
    x_['class']=cls
    sns.scatterplot(data=x_, x='x', y='y', hue='class', ax=ax[0])
    ax[0].set_title('t-SNE', fontweight='bold')

    # LDA
    clf = LinearDiscriminantAnalysis(n_components=2)
    X_t = clf.fit_transform(df, cls)
    x_=pd.DataFrame(X_t,columns=['x','y'])
    x_['class']=cls
    sns.scatterplot(data=x_, x='x', y='y', hue='class', ax=ax[1])
    ax[1].set_title('LDA', fontweight='bold')

    # UMAP
    reducer = umap.UMAP(n_neighbors=10, min_dist=1, n_components=2)
    x_ = reducer.fit_transform(df)
    x_=pd.DataFrame(x_,columns=['x','y'])
    x_['class']=cls
    sns.scatterplot(data=x_, x='x', y='y', hue='class', ax=ax[2])
    _=ax[2].set_title('UMAP', fontweight='bold')

    fig.suptitle(f'Separability of data with respect to {trgt}', fontsize=16)
    plt.tight_layout()


def corel_scatter_(*args,**kwargs):
    df_ = pd.DataFrame(args).T
    r,p=pearsonr(x=df_.iloc[:,0], y=df_.iloc[:,1])
    
    if(r <= -.2):
        c=sns.color_palette()[1]
    elif(-.2 < r and r < .2):
        c=sns.color_palette()[2]
    else:
        c=sns.color_palette()[4]
        
    sns.regplot(data=df_, x=df_.columns[0], y=df_.columns[1],color=c,scatter_kws={'alpha':0.3})
    ax = plt.gca()
    ax.annotate('r = {:.2f}'.format(r), xy=(0.5,0.9), 
                xycoords='axes fraction', ha='center', fontsize=16, fontweight='bold')

def correl_plot(df, color=0):
    df=df.copy()
    df = df.select_dtypes(include='number')
    g = sns.PairGrid(df, height=1.5,aspect=1.5, diag_sharey=False, corner=True)
    g.map_diag(sns.histplot, color=sns.color_palette()[color])
    _=g.map_lower(corel_scatter_)


def qq_plot(data, x, ax):
    
    x_ = np.asarray(data[x])
    osm_uniform = _calc_uniform_order_statistic_medians(len(x_))
    dist = _parse_dist_kw('norm', enforce_subclass=False)
    osm = dist.ppf(osm_uniform)
    osr = np.sort(x_)
    slope, intercept, r, prob, _ = linregress(osm, osr)    
    plot_df = pd.DataFrame({'osm':osm, 'osr':osr,'slope':slope*osm+intercept})
    
    sns.scatterplot(data=plot_df, x='osm', y='osr')
    sns.lineplot(data=plot_df, x='osm', y='slope', color=sns.color_palette()[2])    

    # normality test
    stat, p = shapiro(x_)

    # interpret test statistic
    alpha = 0.05
    if p > alpha:
        normal_test= 'Shapiro-Wilk:\nStatistics=%.3f, \np=%.3f\n(Gaussian)' % (stat, p)#'Sample looks Gaussian (fail to reject H0)'
    else:
        normal_test='Shapiro-Wilk:\nStatistics=%.3f, \np=%.3f\n(NOT Gaussian)' % (stat, p)   
    
    # draw statistic on chart
    xmin, xmax = np.amin(osm), np.amax(osm)
    ymin, ymax = np.amin(x_), np.amax(x_)
    posx = xmin + 0.70 * (xmax - xmin)
    posy = ymin + 0.01 * (ymax - ymin)
    ax.text(posx, posy, normal_test)


def univariate_plot(data, x):

    fig, ax = plt.subplots(2,2, figsize=(12,7))

    # Histogram
    _=sns.histplot(data=data, x=x, ax=ax[0,0])
    ax[0,0].set_title('Histogram', fontweight='bold')

    # Emprirical CDF
    _=sns.ecdfplot(data=data, x=x, ax=ax[1,0])
    _=sns.rugplot(data=data, x=x, ax=ax[1,0], color=sns.color_palette()[2])
    ax[1,0].set_title('Cumulative Distribution Function', fontweight='bold')

    # Boxplot
    _=sns.boxplot(data=data, x=x, ax=ax[0,1], color=sns.color_palette()[2])
    _=ax[0,1].set_title('Boxplot', fontweight='bold')

    # QQ-plot
    _=qq_plot(data=data, x=x, ax=ax[1,1])
    ax[1,1].set_xlabel('Normal theoretical quantiles')
    ax[1,1].set_ylabel('Ordered values')
    ax[1,1].set_title('Data vs normal quantiles', fontweight='bold')

    # Title for plot and layoyt
    fig.suptitle(f'Distribution plots of feature: {x}', fontsize=16)
    plt.tight_layout()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="torgrims",
    version="0.0.12",
    author="Tomas Torgrimsby",
    author_email="torgrimsatpypi@gmail.com",
    packages=["torgrims"],
    description="Package for ML",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://irisml.eu",
    license="OSI Approved (new BSD)",
    python_requires=">=3.6",
    install_requires=[
        "pandas",
        "matplotlib",
        "numpy",
        "scikit-learn",
        "scipy",
        "seaborn",
        "umap-learn",
    ]
)



from .descstat import DescStat
from .helpers import Helpers
from .aml import Aml
from .transformers import ColSelector, TypeCaster, \
    OrdinalEncoder, NanImputer, NullishToNan, OutlierTrimmer, \
    OnehotEncoder, StdScaler, MinMaxScaler, BoxCoxTransformer, \
    PipeMerger 
from .plot import ts_outlier_plot, separability_plot, \
    correl_plot, univariate_plot, qq_plot


import numpy as np
import pandas as pd
from joblib import Parallel
from sklearn.utils.fixes import delayed
from scipy.stats import boxcox
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import FeatureUnion, _fit_transform_one, _transform_one
from pandas.tseries.holiday import USFederalHolidayCalendar
from datetime import datetime
from typing import List, Dict, Union

class ColSelector(BaseEstimator, TransformerMixin):
    '''
    Transformer to select feature from frame
    '''
    def __init__(self, col:Union[str,List[str]])->None:
        super().__init__()
        if type(col)==str:
            col = [col]
        self.col = col

    def fit(self, X:pd.DataFrame, y:pd.DataFrame=None):
        return self

    def transform(self, X:pd.DataFrame, y:pd.DataFrame=None)->pd.DataFrame:
        X=X.copy()
        return X[self.col]

class TypeCaster(BaseEstimator, TransformerMixin):
    '''
    Cast col to proper type, np.nan on error
    '''
    def __init__(self, type:str, date_format:str='%Y-%m-%d')->None:
        self.type=type
        self.date_format=date_format

    def fit(self, X:pd.DataFrame, y:pd.DataFrame=None):
        return self

    def transform(self, X:pd.DataFrame, y:pd.DataFrame=None)->pd.DataFrame:
        X=X.copy()
        for c in X.columns:
            if self.type=='numeric':
                X[c] = pd.to_numeric(X[c], errors='coerce')
            if self.type=='string':
                X[c] = X[c].astype(str)
            if self.type=='date':
                X[c] = pd.to_datetime(X[c], format=self.date_format, errors='coerce')
            return X           

class OrdinalEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, mapping_dict:Dict=None):        
        self.mapping_dict=mapping_dict

    def fit(self, X:pd.DataFrame, y:pd.DataFrame=None):
        X=X.copy()
        if not self.mapping_dict:
            self.mapping_dict={}
            assert X.shape[1]==1, 'Transformer support only one column'
            columns=list(X.columns)
            for c in columns:
                self.mapping_dict[c] = {
                    e[0]:e[1] for e in zip(
                        X[c].unique(),
                        range(len(X[c].unique()))
                    )
                }
        return self

    def transform(self, X:pd.DataFrame, y:pd.DataFrame=None)->pd.DataFrame:
        X = X.copy()
        for k in self.mapping_dict.keys():
            X[k] = X[k].map(self.mapping_dict[k])
        return X

    def inverse_transform(self, X, y=None):
        X = X.copy()
        for v in self.mapping_dict.keys():
            inverse_map = {
                v:k for k,v in self.mapping_dict[v].items()
            }
            X[v] = X[v].map(inverse_map)
        return X        
    
class NanImputer(BaseEstimator, TransformerMixin):
    def __init__(self, strategy:str='mean', fillvalue=np.nan, fillvalue_map:Dict=None):
        #super().__init__()
        self.strategy=strategy
        self.fillvalue=fillvalue
        self.fillvalue_map=fillvalue_map

    def fit(self, X:pd.DataFrame, y:pd.DataFrame=None):
        if not self.fillvalue_map:
            self.fillvalue_map={}
            for c in X.columns:
                if self.strategy == 'mean':    
                    self.fillvalue_map[c]={
                        'fill_value':X[c].mean()
                    }
                if self.strategy == 'median':    
                    self.fillvalue_map[c]={
                        'fill_value':X[c].quantile(.5)
                    }   
                if self.strategy == 'constant':
                    self.fillvalue_map[c]={
                        'fill_value':self.fillvalue
                    }                     
        return self

    def transform(self, X:pd.DataFrame, y:pd.DataFrame=None):
        X=X.copy()
        for c in X.columns:
            X[c]=X[c].replace(np.nan, self.fillvalue_map[c]['fill_value'])
        return X
    
class NullishToNan(BaseEstimator, TransformerMixin):

    NULLISH=[np.nan, 0, -999, 'missing', '?', '', ' ', '  ', '   ']
    def __init__(self, null_def:List[Union[str,int,float]]=NULLISH):
        super().__init__()
        self.null_def=null_def

    def fit(self, X:pd.DataFrame, y:pd.DataFrame=None):
        return self

    def transform(self, X:pd.DataFrame, y:pd.DataFrame=None):
        X=X.copy()
        for c in X.columns:
            X.loc[X[c].isin(self.null_def), c]=np.nan
        return X

class OutlierTrimmer(BaseEstimator, TransformerMixin):
    'replace values of outliers to either bound or np.nan'
    def __init__(self, method='to_bound', threshold_map:Dict=None):
        self.method=method
        self.threshold_map=threshold_map

    def fit(self, X, y=None):
        if not self.threshold_map:
            self.threshold_map={}
            for c in X.columns:
                p75=X[c].quantile(0.75)
                p25=X[c].quantile(0.25)
                lb = p25-1.5*(p75-p25)
                ub = p75+1.5*(p75-p25)                
                self.threshold_map[c]={
                    'lb':lb,
                    'ub':ub
                }
        return self

    def transform(self, X, y=None):
        X=X.copy()
        for c in X.columns:
            lb=self.threshold_map[c]['lb']
            ub=self.threshold_map[c]['ub']

            if self.method == 'to_bound':    
                X[c] = np.where(X[c] < lb, lb, X[c])
                X[c] = np.where(X[c] > ub, ub, X[c])
            else:
                X[c] = np.where(X[c] < lb, np.nan, X[c])
                X[c] = np.where(X[c] > ub, np.nan, X[c])
        return X

class OnehotEncoder(BaseEstimator, TransformerMixin):
    def __init__(self):
        super().__init__()
        self.unique_value_map={}

    def format_colname(self,c):
        return c.lstrip().rstrip().lower().replace('-','_') 

    def fit(self, X:pd.DataFrame, y:pd.DataFrame=None):
        for c in X.columns:
            #self.unique_value_map[c]=[str(e) for e in X[c].unique()]
            self.unique_value_map[c]=list(np.sort(X[c].unique()))
        return self

    def transform(self, X:pd.DataFrame, y:pd.DataFrame=None):
        X = X.copy()
        for k,v in self.unique_value_map.items():
            ohe=pd.DataFrame({
                f'{k}_{self.format_colname(str(e))}':
                    np.where(X[k]==e, 1,0) for e in v
            }, index=X[k].index)
            X=pd.concat([X,ohe], axis=1)
            X.drop(k, axis=1, inplace=True)
        return X

class StdScaler(BaseEstimator, TransformerMixin):
    def __init__(self):
        super().__init__()
        self.scaler_map={}

    def fit(self, X:pd.DataFrame, y:pd.DataFrame=None)->pd.DataFrame:        
        for c in X.columns:
            self.scaler_map[c]={
                'mu':X[c].mean(),
                'std':X[c].std()
            }   
        return self

    def transform(self, X:pd.DataFrame, y:pd.DataFrame=None):
        X = X.copy()
        for c in X.columns:
            mu=self.scaler_map[c]['mu']
            std=self.scaler_map[c]['std']
            X[c]=(X[c]-mu)/std
        return X

class MinMaxScaler(BaseEstimator, TransformerMixin):
    def __init__(self):
        super().__init__()
        self.scaler_map={}

    def fit(self, X:pd.DataFrame, y:pd.DataFrame=None)->pd.DataFrame:        
        for c in X.columns:
            self.scaler_map[c]={
                'min':X[c].min(),
                'max':X[c].max()
            }   
        return self

    def transform(self, X:pd.DataFrame, y:pd.DataFrame=None):
        X = X.copy()
        for c in X.columns:
            min=self.scaler_map[c]['min']
            max=self.scaler_map[c]['max']
            X[c]=(X[c]-min)/(max-min)
        return X

class BoxCoxTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        super().__init__()
        self.scaler_map={}

    def fit(self, X:pd.DataFrame, y:pd.DataFrame=None)->pd.DataFrame:        
        for c in X.columns:
            _,l=boxcox(X[c])
            self.scaler_map[c]={
                'lambda':l
            }   
        return self

    def transform(self, X:pd.DataFrame, y:pd.DataFrame=None):
        X = X.copy() 
        for c in X.columns:    
            l=self.scaler_map[c]['lambda']
            X[c]=boxcox(X[c],l)
        return X


class DaylightHoursTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, earth_axis=23.44, latitude=59.83644689):
        super().__init__()
        self.earth_axis=earth_axis
        self.latitude=latitude

    def daylight_hrs(self,dt_srs):
        
        lat_rad=np.tan(np.radians(self.latitude))
        
        cos_days=np.cos((dt_srs-datetime(2000,1,1)).dt.days * 2 *np.pi / 365.25)
        long_rad=np.tan(np.radians(self.earth_axis)*cos_days)
        
        dd = 1.-long_rad*lat_rad
        hrs=24. * np.degrees(np.arccos(1 - np.clip(dd, 0, 2))) / 180.
        
        return hrs


    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X = X.copy()
        for c in X.columns:
            
            # Calculate dayhours
            X[f'{c}_daylight_hrs']=self.daylight_hrs(X[c])
            X.pop(c)
            
            
        return X
    
    
class DayOfWeekTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        super().__init__()
        self.weekday_map={
                0:'Mon', 
                1:'Tue', 
                2:'Wed', 
                3:'Thu', 
                4:'Fri', 
                5:'Sat', 
                6:'Sun'
            }

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X = X.copy()
        for c in X.columns:
            X[f'{c}_weekday']=X[c].dt.weekday.map(self.weekday_map)
            X.pop(c)         
            
        return X
    
class MonthNumberTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        super().__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X = X.copy()
        for c in X.columns:
            X[f'{c}_month_no']=X[c].dt.month
            X.pop(c)         
            
        return X
    
class QuarterNumberTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        super().__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X = X.copy()
        for c in X.columns:
            X[f'{c}_quarter']=X[c].dt.quarter
            X.pop(c)         
            
        return X
    
    
class IsHolidayTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        super().__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        pass
        X = X.copy()
        
        for c in X.columns:
            mindt=X[c].min().strftime("%Y-%m-%d")
            maxdt=X[c].max().strftime("%Y-%m-%d")
            cal = USFederalHolidayCalendar()
            holiday_flags = pd.Series(1, index=cal.holidays(mindt, maxdt), name=f'{c}_holiday')

            X=X.join(holiday_flags)
            X[f'{c}_holiday'].fillna(0, inplace=True)            
            X.pop(c)
            
        return X


class PipeMerger(FeatureUnion):
    def vstack(self, Xs):
        return pd.concat(Xs, axis="columns", copy=False)

    def lolbua(self, X, y=None, **fit_params):
        for name, trans, weight in self._iter():
            print(f'name is {name}')
            print(f'trans is {trans}')
            print(f'weight is {weight}')
            print('\n\n')

    def fit_transform(self, X, y=None, **fit_params):

        self._validate_transformers()
    
        results = self._parallel_func(X, y, fit_params, _fit_transform_one)
        if not results:
            # All transformers are None
            return np.zeros((X.shape[0], 0))

        Xs, transformers = zip(*results)
        self._update_transformer_list(transformers)

        return self.vstack(Xs)
        
       #if not result:
            # All transformers are None
       #     return np.zeros((X.shape[0], 0))
       # Xs, transformers = zip(*result)
       # self._update_transformer_list(transformers)
       # return self.vstack(Xs)

    def transform(self, X):
        Xs = Parallel(n_jobs=self.n_jobs)(
            delayed(_transform_one)(trans, X, None, weight)
            for name, trans, weight in self._iter())
        if not Xs:
            # All transformers are None
            return np.zeros((X.shape[0], 0))
        Xs = self.vstack(Xs)
        return Xs